from __future__ import division, print_function, absolute_import

from mlpy.agents import Agent
from mlpy.auxiliary.datasets import DataSet
from mlpy.agents.fsm import StateMachine

from naobot.kinematics import NaoMotionController
from .sensors import Sensors, NaoWorldModel
from .vision.image import ImageProcessor


class NaoBot(Agent):
    """The Naobot agent class.

    This class controls the Nao robot.

    Parameters
    ----------
    pip : str
        The IP of the agent for which to process the camera images.
    pport : int
        The port of the agent for which to process the camera images.
    module_type : str
        The agent module type, which is the name of the class.
    task: Task
        The task the agent must complete.
    fsm_filename : str
        Name of the file containing the fsm configuration information.
    keep_history : bool, optional
        If true a history of states and actions performed by the agent
        are recorded and saved to file. Default is False.
    dataset_params : dict
        Non-positional arguments passed to the dataset during instantiation.
    args: tuple
        Positional parameters passed to the agent module.
    kwargs: dict
        Non-positional parameters passed to the agent module.

    """
    @property
    def pip(self):
        ip, _ = self.mid.split(":")
        return ip

    @property
    def pport(self):
        _, port = self.mid.split(":")
        return int(port)

    def __init__(self, pip, pport, module_type, task, fsm_filename, keep_history=False,
                 dataset_params=None, *args, **kwargs):
        assert (isinstance(pip, basestring) and isinstance(pport, int)), \
            "The player IP must be of type string and the player port of type integer."
        super(NaoBot, self).__init__(pip + ":" + str(pport), module_type, task, *args, **kwargs)

        self._keep_history = keep_history

        if self._keep_history:
            dataset_params = dataset_params if dataset_params is not None else {}
            self._history = DataSet(**dataset_params)
            self._history.load()

        self._world_model = NaoWorldModel()
        self._sensors = Sensors(pip, pport)
        self._vision = ImageProcessor(pip, pport)

        self._behavior = StateMachine()
        self._behavior.load_from_file(self, fsm_filename, motion=NaoMotionController(pip, pport))

    def reset(self, t, **kwargs):
        """Reset the naobot's state.

        Parameters
        ----------
        t : float
            The current time (sec).
        kwargs : dict, optional
            Non-positional parameters.

        """
        super(NaoBot, self).reset(t, **kwargs)

        if not self._task.is_complete():
            if self._keep_history:
                self._history.new_sequence()

            self._sensors.reset(t, **kwargs)
            self._vision.reset(t, **kwargs)
            self._behavior.reset(t, **kwargs)

    def enter(self, t):
        """Enter naobot and the agent module.

        Perform initialization tasks here.

        Parameters
        ----------
        t : float
            The current time (sec).

        """
        super(NaoBot, self).enter(t)

        if self._keep_history:
            self._history.new_sequence()

        self._sensors.enter(t)
        self._vision.enter(t)
        self._world_model.enter(t)
        self._behavior.enter(t)

    def update(self, dt):
        """Update naobot and the agent module.

        Naobot and the agent module are updated at every time step
        in the program loop.

        Parameters
        ----------
        dt : float
            The elapsed time (sec)

        """
        super(NaoBot, self).update(dt)

        self._sensors.update(dt)
        self._vision.update(dt)
        self._world_model.update(dt)
        self._behavior.update(dt)

    def exit(self):
        """Exit the agent and the agent module.

        Perform cleanup tasks here.

        """
        super(NaoBot, self).exit()

        self._behavior.exit()
        self._vision.exit()
        self._sensors.exit()
        self._world_model.exit()

    def motion_complete(self):
        """The motion has completed.

        The method is called by the behavior when the motion has completed.

        """
        state = self._task.sensation(**{"state": self._behavior.current_state})
        if state is not None:
            self._module.execute(state)

            if self._keep_history:
                if not self._history.has_field("state"):
                    self._history.add_field("state", len(state), dtype=state.dtype, description=state.description)
                self._history.append("state", state)

                if not self._history.has_field("label"):
                    self._history.add_field("label", 1, dtype=DataSet.DTYPE_OBJECT)
                self._history.append("label", state.name)

        if (self._task.is_episodic and state.is_terminal()) or self._task.termination_requested():
            if self._keep_history:
                self._history.save()

            self._task.terminate(True)
            self._task.request_termination(False)

    def choose_action(self):
        """Determine the next action based on the actor module.

        This method is called by the behavior.

        """
        action = self._module.get_next_action()
        if action is not None:
            self._behavior.post_event(action.name, **{"action": action})

            if self._keep_history:
                if not self._history.has_field("act"):
                    self._history.add_field("act", len(action), dtype=action.dtype, description=action.description)
                self._history.append("act", action)
        else:
            self._behavior.post_event("no-op", **{"delay": self._task.event_delay})
            self._task.request_termination(True)

    def process_motion(self):
        """Process the motion.

        This method is called after updating the state.

        """
        self.motion_complete()

        if self._task.is_complete():
            return

        self.choose_action()



from __future__ import division, print_function, absolute_import

import sys
import traceback

import motion
import almath
from naoqi import ALProxy

from mlpy.modules import Module
from mlpy.auxiliary.misc import listify

from .world_model.model import NaoWorldModel


class NaoMotionController(Module):
    """The controller for Nao's motion.

    An interface to NAOqi's ALMotion [1]_ and ALRobotPosture [2]_ modules
    encapsulating their functionality.

    Parameters
    ----------
    pip : str
        The IP of the agent for which to process the camera images.
    pport : int
        The port of the agent for which to process the camera images.
    default_frame : {FRAME_ROBOT, FRAME_TORSO, FRAME_WORLD}
        The frame to use by default
    use_sensor_values : bool
        Whether to use sensor values or not. Default is True.

    References
    ----------
    .. [1] `ALMotion <http://doc.aldebaran.com/2-1/naoqi/motion/almotion.html>`_
    .. [2] `ALRobotPosture <http://doc.aldebaran.com/2-1/naoqi/motion/alrobotposture.html>`_

    """
    # FRAME
    FRAME_ROBOT = motion.FRAME_ROBOT
    FRAME_TORSO = motion.FRAME_TORSO
    FRAME_WORLD = motion.FRAME_WORLD

    # AXIS MASK
    AXIS_MASK_X = motion.AXIS_MASK_X
    AXIS_MASK_Y = motion.AXIS_MASK_Y
    AXIS_MASK_Z = motion.AXIS_MASK_Z
    AXIS_MASK_WX = motion.AXIS_MASK_WX
    AXIS_MASK_WY = motion.AXIS_MASK_WY
    AXIS_MASK_WZ = motion.AXIS_MASK_WZ
    AXIS_MASK_ALL = motion.AXIS_MASK_ALL
    AXIS_MASK_VEL = motion.AXIS_MASK_VEL
    AXIS_MASK_ROT = motion.AXIS_MASK_ROT

    _effectorNames = ["Head", "LArm", "RArm", "LLeg", "RLeg", "Torso"]
    _jointNames = ["HeadYaw", "HeadPitch", "LShoulderRoll", "LShoulderPitch", "LElbowYaw",
                   "LElbowRoll", "RShoulderRoll", "RShoulderPitch", "RElbowYaw", "RElbowRoll",
                   "LHipYawPitch", "LHipRoll", "LHipPitch", "LKneePitch", "LAnklePitch",
                   "LAnkleRoll", "RHipYawPitch", "RHipRoll", "RHipPitch", "RKneePitch",
                   "RAnklePitch", "RAnkleRoll"]
    _cameraNames = ["CameraTop", "CameraBottom"]

    def __init__(self, pip, pport, default_frame=None, use_sensor_values=None):
        super(NaoMotionController, self).__init__(self._generate_id(pip, pport))

        self._motion_proxy = None
        self._posture_proxy = None

        self._task_id = None

        self._default_frame = default_frame if default_frame is not None else NaoMotionController.FRAME_ROBOT
        self._use_sensor_values = use_sensor_values if use_sensor_values is not None else True

        self._initialize(pip, pport)

        robot_config = self._motion_proxy.getRobotConfig()
        NaoWorldModel().set_robot_params(robot_config[1][0], robot_config[1][2])

        self._effector = NaoWorldModel().get_positions("effector")
        self._joint = NaoWorldModel().get_positions("joint")
        self._camera = NaoWorldModel().get_positions("camera")

        self._init_effector()
        self._init_joint()
        self._init_camera()

    def reset(self, t, **kwargs):
        """Resets the motion controller.

        Ensure that all proxies are valid.

        Parameters
        ----------
        t : float
            The current time (sec)
        kwargs : dict, optional
            Non-positional parameters, optional.

        """
        super(NaoMotionController, self).reset(t, **kwargs)
        self._initialize(self.mid.split(':')[1], int(self.mid.split(':')[2]))

    def enter(self, t):
        """Perform preliminary setup.

        Parameters
        ----------
        t : float
            The current time (sec).

        """
        super(NaoMotionController, self).enter(t)

        self.go_to_default_posture()
        self.update(0.0)

    def update(self, dt):
        """Update the motion controller.

        Read the positions and rotation angles for all effectors,
        joints, and cameras.

        Parameters
        ----------
        dt : float
            The elapsed time (sec)

        """
        super(NaoMotionController, self).update(dt)

        self._update_effector()
        self._update_joint()
        self._update_camera()

    def go_to_posture(self, name, speed):
        """Go to predefined posture.

        This encapsulates the function `goToPosture
        <http://doc.aldebaran.com/2-1/naoqi/motion/alrobotposture-api.html#ALRobotPostureProxy::goToPosture__ssC.floatC>`_
        from the ALRobotPosture module.

        Parameters
        ----------
        name : str
            The name of the posture.
        speed : float
            Relative speed between 0.0 and 1.0.

        Notes
        -----
        This is a blocking call.

        """
        self._posture_proxy.goToPosture(name, speed)

    def go_to_default_posture(self):
        """Go to default posture.

        This encapsulates the function `goToPosture
        <http://doc.aldebaran.com/2-1/naoqi/motion/alrobotposture-api.html#ALRobotPostureProxy::goToPosture__ssC.floatC>`_
        from the ALRobotPosture module.

        The default posture is `StandInit` and the agent
        transitions to this posture with a speed of 0.5.

        """
        self._posture_proxy.goToPosture("StandInit", 0.5)

    def interpolate_angles(self, joints, angles, times, is_absolute=True):
        """Interpolates joints to a target angles.

        Interpolates one or more joints to a target angle or along timed trajectories.

        This encapsulates the function `angleInterpolation
        <http://doc.aldebaran.com/2-1/naoqi/motion/control-joint-api.html#ALMotionProxy::angleInterpolation__AL::ALValueCR.AL::ALValueCR.AL::ALValueCR.bCR>`_
        from the ALMotion module.

        Parameters
        ----------
        joints : str or list[str]
            The name(s) of the joints or chains.
        angles : float or list[float] or list[list[float]]
            The target angle(s) in radians the joints are interpolated to.
        times : float or list[float] or list[list[float]]
            The time in seconds for the interpolation.
        is_absolute : bool, optional
            If true, the movement is described in absolute angles, else the
            angles are relative to the current angels. Default is True.

        Notes
        -----
        This is a non-blocking call.

        """
        try:
            self._task_id = self._motion_proxy.post.angleInterpolation(joints, angles, times, is_absolute)
        except:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            traceback.print_exception(exc_type, exc_value, exc_traceback)
            sys.exit(1)

    def interpolate_positions(self, effectors, delta_pos, times, frames=None, axis_masks=None):
        """Interpolate end-effectors to target positions and orientations.

        Moves end-effectors to the given position and orientation over time.

        This encapsulates the function `positionInterpolation
        <http://doc.aldebaran.com/2-1/naoqi/motion/control-cartesian-api.html#almotionproxy-positioninterpolations1>`_
        from the ALMotion module.

        Parameters
        ----------
        effectors : str or list[str]
            List of effector names. Effector name could be "Torso" or chain name.
        delta_pos : ndarray[float] or list[ndarray[float]] or list[list[ndarray[float]]]
            The relative position changes for each effector.
        times : list[float] or list[list[float]]
            Vector of times in seconds corresponding to the path points.
        frames : list[int], optional
            List of task frames {FRAME_TORSO = 0, FRAME_WORLD = 1, FRAME_ROBOT = 2 }.
            Default is FRAME_ROBOT.
        axis_masks : list[int]
            List of axis masks. True for axes that should be controlled. Default is
            AXIS_MASK_ALL.

        Notes
        -----
        This is a non-blocking call.

        """
        effectors = listify(effectors)
        n = len(effectors)

        frames = listify(frames) if frames is not None else [NaoMotionController.FRAME_ROBOT] * n
        assert (len(frames) == n), "effectors and frames must have the same number of elements"
        axis_masks = listify(axis_masks) if axis_masks is not None else [NaoMotionController.AXIS_MASK_ALL] * n
        assert (len(axis_masks) == n), "effectors and axis_masks must have the same number of elements"

        try:
            path = []
            for e, d, f in zip(effectors, listify(delta_pos), frames):
                path.append(self._calculate_target_pos(e, d, f))
            self._task_id = self._motion_proxy.post.positionInterpolations(effectors, frames, path, axis_masks, times)
        except:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            traceback.print_exception(exc_type, exc_value, exc_traceback)
            sys.exit(1)

    def transform(self, effectors, delta_tf, times, frames=None, axis_masks=None):
        """Moves end-effector to the given transforms over time.

        This encapsulates the function `transformInterpolation
        <http://doc.aldebaran.com/2-1/naoqi/motion/control-cartesian-api.html#almotionproxy-transforminterpolations1>`_
        from the ALMotion module.

        Parameters
        ----------
        effectors : str or list[str]
            List of effector names. Effector name could be "Torso" or chain name.
        delta_tf : ndarray[float] or list[ndarray[float]] or list[list[ndarray[float]]]
            The relative position changes.
        times : list[float] or list[list[float]]
            List of times in seconds corresponding to the path points.
        frames : list[int], optional
            List of task frames {FRAME_TORSO = 0, FRAME_WORLD = 1, FRAME_ROBOT = 2 }.
            Default is FRAME_ROBOT.
        axis_masks : list[int], optional
            List of axis masks. True for axes that should be controlled. Default is
            AXIS_MASK_ALL.

        Notes
        -----
        This is a non-blocking call.

        """
        effectors = listify(effectors)
        n = len(effectors)

        frames = listify(frames) if frames is not None else [NaoMotionController.FRAME_ROBOT] * n
        assert (len(frames) == n), "effectors and frames must have the same number of elements"
        axis_masks = listify(axis_masks) if axis_masks is not None else [NaoMotionController.AXIS_MASK_ALL] * n
        assert (len(axis_masks) == n), "effectors and axis_masks must have the same number of elements"

        try:
            path = []
            for e, d, f in zip(effectors, listify(delta_tf), frames):
                path.append(self._calculate_target_tf(e, d, f))
            self._task_id = self._motion_proxy.post.transformInterpolations(effectors, frames, path, axis_masks, times)
        except:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            traceback.print_exception(exc_type, exc_value, exc_traceback)
            sys.exit(1)

    def move_joint(self, name, angle, fraction_max_speed=1.0):
        """ Sets the joint angle directly.

        This encapsulates the function `setAngles
        <http://doc.aldebaran.com/2-1/naoqi/motion/control-joint-api.html#ALMotionProxy::setAngles__AL::ALValueCR.AL::ALValueCR.floatCR>`_
        from the ALMotion module.

        Parameters
        ----------
        name: str | list[str]
            The name(s) of the joints or chains.
        angle: float | list[floats]
            The angles in radians to which the angle is set.
        fraction_max_speed: float, optional
            The fraction of maximum speed to use. Default is 1.0.

        Notes
        -----
        This is a non-blocking call.

        """
        self._motion_proxy.setAngles(name, angle, fraction_max_speed)

    def wait(self):
        """Waits until the end of the current task"""
        self._motion_proxy.wait(self._task_id, 0)

    def is_running(self):
        """Check if the current task is still running."""
        if not self._task_id:
            return False
        return self._motion_proxy.isRunning(self._task_id)

    def _initialize(self, pip, pport):
        """Initializes the motion controller

        Ensure that valid proxies to the Nao motion and robot
        posture exist.

        Parameters
        ----------
        pip : str
            The IP of the agent for which to process the camera images.
        pport : int
            The port of the agent for which to process the camera images.

        """
        try:
            # noinspection PyBroadException
            try:
                self._motion_proxy.ping()
                self._posture_proxy.ping()
            except:
                self._motion_proxy = ALProxy("ALMotion", pip, pport)
                self._posture_proxy = ALProxy("ALRobotPosture", pip, pport)

                self._motion_proxy.wakeUp()
        except:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            traceback.print_exception(exc_type, exc_value, exc_traceback)
            sys.exit(1)

    def _init_effector(self):
        """Initialize the position and orientation of all effectors."""
        for x in self._effectorNames:
            self._effector[x] = 0.0

    def _init_joint(self):
        """Initialize the position and orientation of all joints."""
        for x in self._jointNames:
            self._joint[x] = 0.0

    def _init_camera(self):
        """Initialize the position and orientation of all cameras."""
        for x in self._cameraNames:
            self._camera[x] = 0.0

    def _update_effector(self):
        """Update the position and orientation of all effectors."""
        for x in self._effectorNames:
            self._effector[x] = self._motion_proxy.getPosition(x, self._default_frame, self._use_sensor_values)

    def _update_joint(self):
        """Update the position and orientation of all joints."""
        for x in self._jointNames:
            self._joint[x] = self._motion_proxy.getPosition(x, self._default_frame, self._use_sensor_values)

    def _update_camera(self):
        """Update the position and orientation of all cameras."""
        for x in self._cameraNames:
            self._camera[x] = self._motion_proxy.getPosition(x, self._default_frame, self._use_sensor_values)

    def _calculate_target_tf(self, effector, delta_tf, frame):
        """Calculate the target transform.

        Calculate the target transform from the current position of the effector
        and the relative position change.

        Parameters
        ----------
        effector : str
            String of effector name.
        delta_tf : ndarray[float]
            The relative position change.
        frame : {FRAME_TORSO, FRAME_WORLD, FRAME_ROBOT}
            The task frame .

        """
        init_tf = almath.Transform(self._motion_proxy.getTransform(effector, frame, self._use_sensor_values))

        delta_tf = listify(delta_tf)
        d = delta_tf.pop(0)

        target_tf = init_tf * almath.Transform().fromPosition(*d)
        path = list(target_tf.toVector())

        if delta_tf:
            path = [path]
            for d in delta_tf:
                target_tf *= almath.Transform().fromPosition(*d)
                path.append(list(target_tf.toVector()))
        return path

    def _calculate_target_pos(self, effector, delta_pos, frame):
        """Calculate the target position.

        Calculate the target position from the current position of the effector
        and the relative position change.

        Parameters
        ----------
        effector : str
            String of effector name.
        delta_pos : ndarray[float]
            The relative position change.
        frame : {FRAME_TORSO, FRAME_WORLD, FRAME_ROBOT}
            The task frame .

        """
        init_pos = self._motion_proxy.getPosition(effector, frame, self._use_sensor_values)

        delta_pos = listify(delta_pos)
        d = delta_pos.pop(0)

        target_pos = almath.Position6D(init_pos)
        target_pos += almath.Position6D(d)
        path = list(target_pos.toVector())

        if delta_pos:
            path = [path]
            for d in delta_pos:
                target_pos += almath.Position6D(d)
                path.append(list(target_pos.toVector()))
        return path

    def _generate_id(self, pip, pport):
        return "%s.%s:%s:%i:%i" % (self.__class__.__module__, self.__class__.__name__, pip, pport, next(self._ids))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_naobot
----------------------------------

Tests for `naobot` module.
"""


class TestNaobot(object):

    def set_method(self):
        pass

    def test_something(self):
        pass

    def teardown_method(self, _):
        pass



# -*- coding: utf-8 -*-
from __future__ import division, print_function, absolute_import

__all__ = ['agent', 'behavior', 'environment', 'kinematics', 'sensors', 'world_model', 'vision']

__author__ = 'Astrid Jackson'
__email__ = 'ajackons@eecs.ucf.edu'
__version__ = '0.1.0'



#!/usr/bin/env python
# -*- coding: utf-8 -*-


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read().replace('.. :changelog:', '')

requirements = [
    'numpy',
    'Pillow',
    'mlpy',
]

test_requirements = [
    'pytest',
]

import naobot

setup(
    name='naobot',
    version=naobot.__version__,
    description="Control of a Nao robot",
    long_description=readme + '\n\n' + history,
    author="Astrid Jackson",
    author_email='ajackons@eecs.ucf.edu',
    url='https://readthedocs.org/builds/naobot',
    download_url='https://github.com/evenmarbles/naobot',
    packages=[
        'naobot',
    ],
    package_dir={'naobot':
                 'naobot'},
    include_package_data=True,
    install_requires=requirements,
    license="MIT",
    zip_safe=False,
    keywords='naobot',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
    ],
    test_suite='tests',
    tests_require=test_requirements
)



from __future__ import division, print_function, absolute_import

import sys
import traceback

from naoqi import ALProxy

from mlpy.modules import Module

from .world_model.model import NaoWorldModel


class Sensors(Module):
    """Nao's sensor class.

    An interface to NAOqi's ALMemory [1]_ module encapsulating the functionality.

    This class reads the joint angles, the inertial information, the force
    sensitive resistance (FSR), and the bumper values from the Nao's memory at
    each time step.

    Parameters
    ----------
    pip : str
        The IP of the agent for which to process the camera images.
    pport : int
        The port of the agent for which to process the camera images.

    Notes
    -----
    The joint angles are measured in radians. And the FSR are measured in kilogram
    at the feet. Bumpers return a 2-state float value 0.0 (unpressed) or 1.0 (pressed).

    The inertial sensors supported are:

        Gyroscope
            Extract the direct rotation speed values in rad/s.
        Accelerometer
            3-axis acceleration in m/s.
        Angles
            Three inclination angles in radians of the robot body are computed.

    For more information see [2]_.

    References
    ----------
    .. [1] `ALMemory <http://doc.aldebaran.com/2-1/naoqi/core/almemory.html>`_
    .. [2] `NAO - Actuator & Sensor list <http://doc.aldebaran.com/2-1/family/nao_dcm/actuator_sensor_names.html>`_

    """
    _jointNames = ["HeadYaw", "HeadPitch", "LShoulderRoll", "LShoulderPitch", "LElbowYaw",
                   "LElbowRoll", "RShoulderRoll", "RShoulderPitch", "RElbowYaw", "RElbowRoll",
                   "LHipYawPitch", "LHipRoll", "LHipPitch", "LKneePitch", "LAnklePitch",
                   "LAnkleRoll", "RHipYawPitch", "RHipRoll", "RHipPitch", "RKneePitch",
                   "RAnklePitch", "RAnkleRoll"]
    _inertialNames = ["AccX", "AccY", "AccZ", "GyroscopeX", "GyroscopeY", "AngleX", "AngleY"]
    _footNames = ["LFoot/FSR/FrontLeft", "LFoot/FSR/FrontRight", "LFoot/FSR/RearLeft", "LFoot/FSR/RearRight",
                  "RFoot/FSR/FrontLeft", "RFoot/FSR/FrontRight", "RFoot/FSR/RearLeft", "RFoot/FSR/RearRight"]
    _bumperNames = ["LFoot/Bumper/Left", "LFoot/Bumper/Right", "RFoot/Bumper/Left", "RFoot/Bumper/Right"]

    def __init__(self, pip, pport):
        super(Sensors, self).__init__(self._generate_id(pip, pport))

        self._memoryProxy = None
        self._initialize(pip, pport)

        self._joint = NaoWorldModel().get_sensors("joint")
        self._inertial = NaoWorldModel().get_sensors("inertial")
        self._foot = NaoWorldModel().get_sensors("foot")
        self._bumper = NaoWorldModel().get_sensors("bumper")

        self._init_joint()
        self._init_inertial()
        self._init_foot()
        self._init_bumper()

    def reset(self, t, **kwargs):
        """Resets Nao's sensors.

        Ensure that all proxies are valid.

        Parameters
        ----------
        t : float
            The current time (sec)
        kwargs : dict, optional
            Non-positional parameters, optional.

        """
        super(Sensors, self).reset(t, **kwargs)
        self._initialize(self.mid.split(':')[1], int(self.mid.split(':')[2]))

    def enter(self, t):
        """Perform preliminary setup.

        Parameters
        ----------
        t : float
            The current time (sec).

        """
        super(Sensors, self).enter(t)
        self.update(0.0)

    def update(self, dt):
        """Update Nao's sensor information.

        Read the positions and rotation angles for all joints, inertial
        sensors, foot sensors, and bumpers. Also check if the robot
        has fallen.

        Parameters
        ----------
        dt : float
            The elapsed time (sec)

        """
        super(Sensors, self).update(dt)

        self._has_fallen()
        self._update_joint()
        self._update_inertial()
        self._update_foot()
        self._update_bumper()

    def _initialize(self, pip, pport):
        """Initializes Nao's sensor.

        Ensure that valid proxies to the Nao memory exist.

        Parameters
        ----------
        pip : str
            The IP of the agent for which to process the camera images.
        pport : int
            The port of the agent for which to process the camera images.

        """
        try:
            try:
                self._memoryProxy.ping()
            except:
                self._memoryProxy = ALProxy("ALMemory", pip, pport)
        except:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            traceback.print_exception(exc_type, exc_value, exc_traceback)
            sys.exit(1)

    def _has_fallen(self):
        """Identify if the robot has fallen."""
        NaoWorldModel().has_fallen = self._memoryProxy.getData("footContact") == 0.0

    def _init_joint(self):
        """Initialize all joint angles."""
        for x in self._jointNames:
            self._joint[x] = 0.0

    def _init_inertial(self):
        """Initialize all inertial information."""
        for x in self._inertialNames:
            self._inertial[x] = 0.0

    def _init_foot(self):
        """Initialize all FSR values."""
        for x in self._footNames:
            self._foot[x] = 0.0

    def _init_bumper(self):
        """Initialize all bumper values."""
        for x in self._bumperNames:
            self._bumper[x] = 0.0

    def _update_joint(self):
        """Update the angle of all joints.

        Reads the joint angles in radians.

        """
        for x in self._jointNames:
            self._joint[x] = self._memoryProxy.getData("Device/SubDeviceList/" + x + "/Position/Sensor/Value")

    def _update_inertial(self):
        """Extract information from the inertial sensor in the center of the body.

        Gyroscope
            Extract the direct rotation speed values in rad/s.
        Accelerometer
            3-axis acceleration in m/s.
        Angles
            Three inclination angles in radians of the robot body are computed.

        """
        for x in self._inertialNames:
            self._inertial[x] = self._memoryProxy.getData("Device/SubDeviceList/InertialSensor/" + x + "/Sensor/Value")

    def _update_foot(self):
        """Update the force sensitive resistance (FSR) of all foot sensors.

        These sensors measure a resistance change according to the pressure applied.
        The measurements are in kilograms.

        """
        for x in self._footNames:
            self._foot[x] = self._memoryProxy.getData("Device/SubDeviceList/" + x + "/Sensor/Value")

    def _update_bumper(self):
        """Update all bumpers.

        This identifies whether the bumper was pressed or not.

        """
        for x in self._bumperNames:
            self._bumper[x] = self._memoryProxy.getData("Device/SubDeviceList/" + x + "/Sensor/Value")

    def _generate_id(self, pip, pport):
        return "%s.%s:%s:%i:%i" % (self.__class__.__module__, self.__class__.__name__, pip, pport, next(self._ids))



from __future__ import division, print_function, absolute_import

import sys
import traceback
import importlib

from mlpy.agents.fsm import StateMachine, FSMState
from mlpy.auxiliary.io import import_module_from_path

from .world_model.model import NaoWorldModel
from .kinematics import NaoMotionController


class NaoState(FSMState):
    """FSM base state for all bot states.

    Parameters
    ----------
    motion : NaoMotionController
        The motion controller.

    """

    # noinspection PyArgumentList
    def __init__(self, motion):
        super(NaoState, self).__init__()

        self._motion = motion

    def reset(self, t, **kwargs):
        """Resets Nao's sensors.

        Ensure that all proxies are valid.

        Parameters
        ----------
        t : float
            The current time (sec)
        kwargs : dict, optional
            Non-positional parameters, optional.

        """
        super(NaoState, self).reset(t, **kwargs)

        if self._motion:
            self._motion.reset(t, **kwargs)

    def update(self, dt):
        """Update the state.

        Update the state and handle state transitions based on events.

        Parameters
        ----------
        dt : float
            The elapsed time (sec)

        Returns
        -------
        Event
            The transition event.

        """
        super(NaoState, self).update(dt)

        self._motion.update(dt)

        if NaoWorldModel().has_fallen:
            return "has_fallen"


class NaoBehavior(StateMachine):
    """The finite state machine for the soccer bot behavior.

    This class controls Nao's behavior based on state transitions
    specified in the behavior configuration.

    Parameters
    ----------
    pip : str
        The IP of the agent for which to process the camera images.
    pport : int
        The port of the agent for which to process the camera images.
    states : FSMState | list[FSMState], optional
        A list of states.
    initial : str, optional
        The initial state.
    transitions : list[dict] | list[list], optional
        Transition information.
    onupdate : list[dict] | list[list], optional
        Callback information to be executed on update.
    model : object, optional
        Reference to the object owning the FSM.
    configuration : ConfigMgr, optional
        The configuration manager containing the information
        for setting up the FSM. This allowing a data driven
        FSM setup.

    """
    def __init__(self, pip, pport, states=None, initial=None, transitions=None,
                 onupdate=None, model=None, configuration=None):
        super(NaoBehavior, self).__init__(states, initial, transitions, onupdate)
        self._load_fsm(model, NaoMotionController(pip, pport), configuration)

    # noinspection PyUnresolvedReferences
    def _load_fsm(self, model, motion, cfg):
        """Load the FSM from the configuration file.

        Parameters
        ----------
        model: object
            Reference to the object owning the FSM.
        motion: NaoMotionController
            The motion controller, controlling Nao's motions.
        cfg: ConfigMgr
            The configuration manager containing the information
            for setting up the FSM. This allowing a data driven
            FSM setup.

        """
        try:
            module = import_module_from_path(cfg.get("Module"), "task_fsm")

            self._initial = cfg.get("Initial")

            for state in cfg.get("States"):
                kwargs = {}
                if isinstance(state, dict):
                    kwargs = state.itervalues().next()
                    state = state.iterkeys().next()
                state_c = getattr(module, state)(motion, **kwargs)

                self.add_states(state_c)

            for t in cfg.get("Transitions"):
                for key in t.iterkeys():
                    c = None
                    if key == "before":
                        c = "before"
                    if key == "after":
                        c = "after"
                    if key == "conditions":
                        c = "conditions"

                    if c is not None:
                        if c in ["before", "after"]:
                            if t[c]["model"] == model.__class__.__name__:
                                t[c] = getattr(model, t[c]["func"])
                            else:
                                k = t[c]["model"].rfind(".")
                                n = t[c]["model"][:k]
                                m = t[c]["model"][k+1:]
                                # noinspection PyUnusedLocal
                                module = importlib.import_module(n)
                                t[c]["model"] = eval("module." + m)
                        if c == "conditions":
                            for i, cond in enumerate(t[c]):
                                t[c][i] = eval(t[c][i])

                self.add_transition(t["event"], t["source"], t["dest"],
                                    conditions=t["conditions"] if "conditions" in t else None,
                                    before=t["before"] if "before" in t else None,
                                    after=t["after"] if "after" in t else None)

            for u in cfg.get("OnUpdate"):
                for key in u.iterkeys():
                    c = None
                    if key == "onupdate":
                        c = "onupdate"
                    if key == "conditions":
                        c = "conditions"

                    if c is not None:
                        if c in "onupdate":
                            if u[c]["model"] == model.__class__.__name__:
                                u[c] = getattr(model, u[c]["func"])
                            else:
                                k = u[c]["model"].rfind(".")
                                n = u[c]["model"][:k]
                                m = u[c]["model"][k+1:]
                                # noinspection PyUnusedLocal
                                module = importlib.import_module(n)
                                u[c]["model"] = eval("module." + m)
                        if c == "conditions":
                            for i, cond in enumerate(u[c]):
                                u[c][i] = eval(u[c][i])

                self.add_onupdate(u["source"], onupdate=u["onupdate"] if "onupdate" in u else None,
                                  conditions=u["conditions"] if "conditions" in u else None)
        except KeyError:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            traceback.print_exception(exc_type, exc_value, exc_traceback)
            sys.exit(1)
        except:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            traceback.print_exception(exc_type, exc_value, exc_traceback)
            sys.exit(1)
